(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Evaluator4.sml
 *  o Description: .
 *)


structure Evaluator4 = struct

open EvaluatorsBackground

val empty_env = mk_term ("env", []) []
(* NOTE: Where is the env comming from? *)

fun destruct_token_parameter (value, kind) =
    if kind = "token"
    then value
    else raise FAIL

fun mk_token_parm opid = (opid, "token")

fun mk_simple_term_op st opid terms =
    mk_term (st, [mk_token_parm opid])
	    (map (fn term => ([], term)) terms)

fun mk_term_op st opid terms =
    mk_term (st, [mk_token_parm opid]) terms

fun dest_inter_2 term =
    let val ((opid, [param]),
	     [([], t),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, t, env, k)
    end handle _ => raise FAIL

val dest_E2 = dest_inter_2
val dest_CAN2 = dest_inter_2

fun dest_inter_2_2 term =
    let val ((opid, [param]),
	     [([], t),
	      ([], t1),
	      ([], t2),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, t, t1, t2, env, k)
    end handle _ => raise FAIL

val dest_C2 = dest_inter_2_2

(* Destruct the final F3 in each of Evaluator3's branches *)
fun dest_final_2 term =
    let val ((opid, [param]),
	     [([], v),
	      ([], e),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, v, e, env, k)
    end handle _ => raise FAIL

val dest_DoOp = dest_final_2
val dest_DoCan = dest_final_2

(* Same as dest_final_2 but the second 2 means that it also has 2
 * extra terms corresponding to two different calls to Eval. *)
fun dest_final_2_2 term =
    let val ((opid, [param]),
	     [([], v),
	      ([], e),
	      ([], t1),
	      ([], t2),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, v, e, t1, t2, env, k)
    end handle _ => raise FAIL

val dest_DoCmp = dest_final_2_2

(* Destruct the final F2 in each of Evaluator3's branches *)
fun dest_final_1 term =
    let val ((opid, [param]),
	     [([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, env, k)
    end handle _ => raise FAIL

val dest_DoMinus = dest_final_1
val dest_DoIn = dest_final_1

(* Same as dest_final_1 but the second 2 means that it also has 2
 * extra terms corresponding to two different calls to Eval. *)
fun dest_final_1_2 term =
    let val ((opid, [param]),
	     [([], t1),
	      ([], t2),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, t1, t2, env, k)
    end handle _ => raise FAIL

val dest_DoTest = dest_final_1_2

fun dest_AE2 term =
    let val ((opid, [param]),
	     [([], t),
	      ([], t0),
	      ([], t1),
	      ([], t2),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, t, t0, t1, t2, env, k)
    end handle _ => raise FAIL

fun dest_DoAtomEq term =
    let val ((opid, [param]),
	     [([], v),
	      ([], e),
	      ([], t0),
	      ([], t1),
	      ([], t2),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, v, e, t0, t1, t2, env, k)
    end handle _ => raise FAIL

fun dest_DoSpread term =
    let val ((opid, [param]),
	     [([x, y], B),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([x, y], B), env, k)
    end handle _ => raise FAIL

fun dest_DoDecide term =
    let val ((opid, [param]),
	     [([x], A),
	      ([y], B),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([x], A), ([y], B), env, k)
    end handle _ => raise FAIL

fun dest_DoApply term =
    let val ((opid, [param]),
	     [([], arg),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([], arg), env, k)
    end handle _ => raise FAIL

fun dest_DoCallbyvalue term =
    let val ((opid, [param]),
	     [([x], B),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([x], B), env, k)
    end handle _ => raise FAIL

fun dest_DoCallbyvalueall term =
    let val ((opid, [param]),
	     [([x], B),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([x], B), env, k)
    end handle _ => raise FAIL

fun dest_DoListInd term =
    let val ((opid, [param]),
	     [([], nilcase),
	      ([x, xs, r], conscase),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([], nilcase), ([x, xs, r], conscase), env, k)
    end handle _ => raise FAIL

fun dest_DoInd term =
    let val ((opid, [param]),
	     [([x, rd], downcase),
	      ([], basecase),
	      ([y, ru], upcase),
	      ([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, ([x, rd], downcase), ([], basecase), ([y, ru], upcase), env, k)
    end handle _ => raise FAIL

fun dest_DoLTC term =
    let val ((opid, [param]),
	     [([], env),
	      ([], k)]) = dest_term term
    in (destruct_token_parameter param, env, k)
    end handle _ => raise FAIL

val halt_term = mk_simple_term "Halt" []

(* EVALUATOR4 - defunctionalization *)
fun Evaluator4' steps cbv_all term env k =
    let fun ApplyCont c (v, e, num) =
	    let val name = opid_of_term c
	    in if name = "Halt"
	       then (closure2term (v, e), steps - num)
	       else if name = "E2"
	       then let val (opid, t2, env, k) = dest_E2 c
		    in Eval num t2 env (mk_simple_term_op "DoOp" opid [v, e, env, k])
		    end
	       else if name = "DoOp"
	       then let val (opid, v1, e1, env, k) = dest_DoOp c
		    in let val steps3 = decr_steps num
		       in ApplyCont k (do_primitive_int_op opid v1 v, empty_env, steps3)
		       end
		       handle _ => ApplyCont k (mk_simple_term opid [closure2term (v1, e1),
								     closure2term (v, e)],
						env,
						num)
		    end
	       else if name = "DoMinus"
	       then let val (opid, env, k) = dest_DoMinus c
		    in if is_term "natural_number" v
		       then ApplyCont k (mk_simple_term opid [v], e, num)
		       else let val steps2 = decr_steps num
			    in ApplyCont k (do_primitive_minus v , e, steps2)
			    end
			    handle _ => ApplyCont k (mk_simple_term opid [v], env, num)
		    end
	       else if name = "C2"
	       then let val (opid, t2, t3, t4, env, k) = dest_C2 c
		    in Eval num t2 env (mk_simple_term_op "DoCmp" opid [v, e, t3, t4, env, k])
		    end
	       else if name = "DoCmp"
	       then let val (opid, v1, env1, t3, t4, env, k) = dest_DoCmp c
		    in let val steps3 = decr_steps num
		       in if do_primitive_cmp opid v1 v
			  then Eval steps3 t3 env k
			  else Eval steps3 t4 env k
		       end handle _ => ApplyCont k (mk_simple_term opid [closure2term (v1, env1),
									 closure2term (v, e),
									 t3,
									 t4],
						    env,
						    num)
		    end
	       else if name = "AE2"
	       then let val (opid, t2, t0, t3, t4, env, k) = dest_AE2 c
		    in Eval num t2 env (mk_simple_term_op "DoAtomEq" opid [v, e, t0, t3, t4, env, k])
		    end
	       else if name = "DoAtomEq"
	       then let val (opid, v1, env1, t0, t3, t4, env, k) = dest_DoAtomEq c
		    in let val steps3 = decr_steps num
			   val n = (firstnat t0) handle _ => 0
		       in if compare_atomn n v1 v
			  then Eval steps3 t3 env k
			  else Eval steps3 t4 env k
		       end handle _ => ApplyCont k (mk_term (opid, []) [([], closure2term (v1, env1)),
									([], closure2term (v, e)),
									([], t3),
									([], t4)],
						    env,
						    num)
		    end
	       else if name = "DoTest"
	       then let val (opid, t2, t3, env, k) = dest_DoTest c
		    in let val steps2 = decr_steps num
		       in if do_primitive_test opid v
			  then Eval steps2 t2 env k
			  else Eval steps2 t3 env k
		       end handle _ => ApplyCont k (mk_simple_term opid [closure2term (v, e),
									 t2,
									 t3],
						    env,
						    num)
		    end
	       else if name = "DoSpread"
	       then let val (opid, ([x, y], B), env, k) = dest_DoSpread c
		    in let val steps2 = decr_steps num
			   val (a, b) = dest_pair v
			   val env2 = add_env_binding (add_env_binding env x a e) y b e
		       in Eval steps2 B env2 k
		       end handle _ => ApplyCont k (mk_term (opid, []) [([], closure2term (v, e)),
									([x, y], B)],
						    env,
						    num)
		    end
	       else if name = "DoDecide"
	       then let val (opid, ([x], A), ([y], B), env, k) = dest_DoDecide c
		    in let val steps2 = decr_steps num
		       in if is_term "inl" v
			  then let val env2 = add_env_binding env x (subtermn 1 v) e
			       in Eval steps2 A env2 k
			       end
			  else if is_term "inr" v
			  then let val env2 = add_env_binding env y (subtermn 1 v) e
			       in Eval steps2 B env2 k
			       end
			  else raise FAIL
		       end handle _ => ApplyCont k (mk_term (opid, []) [([], closure2term (v, e)),
									([x], A),
									([y], B)],
						    env,
						    num)
		    end
	       else if name = "DoApply"
	       then let val (opid, ([], arg), env, k) = dest_DoApply c
		    in let val steps2 = decr_steps num
			   val (x, B) = dest_lambda v
			   val env' = add_env_binding e x arg env
		       in Eval steps2 B env' k
		       end handle _ => ApplyCont k (mk_apply_term (closure2term (v, e)) arg, env, num)
		    end
	       else if name = "DoCallbyvalue"
	       then let val (opid, ([x], B), env, k) = dest_DoCallbyvalue c
		    in let val steps2 = decr_steps num
		       in if is_primitive_value v
			  then let val env' = add_env_binding env x v e
			       in Eval steps2 B env' k
			       end
			  else raise FAIL
		       end handle _ => ApplyCont k (mk_term (opid, []) [([], closure2term (v, e)),
									([x], B)],
						    env,
						    num)
		    end
	       else if name = "DoCallbyvalueall"
	       then let val (opid, ([x], B), env, k) = dest_DoCallbyvalueall c
		    in let val steps2 = decr_steps num
		       in if is_complete_primitive_value v
			  then let val env' = add_env_binding env x v e
			       in Evaluator4' steps2 cbv_all B env' k
			       end
			  else raise FAIL
		       end handle _ => ApplyCont k (mk_term (opid, []) [([], closure2term (v, e)),
									([x], B)],
						    env,
						    num)
		    end
	       else if name = "DoListInd"
	       then let val (opid, ([], nilcase), ([x, xs, r], conscase), env, k) = dest_DoListInd c
		    in let val steps2 = decr_steps num
		       in if is_term "nil" v
			  then Eval steps2 nilcase env k
			  else if is_term "cons" v
			  then let val va = subtermn 1 v
				   val vb = subtermn 2 v
				   val env2 = add_env_binding env xs vb e
				   val env3 = add_env_binding env2 x va e
				   val w    = fresh_variable term
				   val env4 = add_env_binding env w vb e
				   val t2 = mk_term (opid, []) [([],         mk_variable_term w),
								([],         nilcase),
								([x, xs, r], conscase)]
				   val env4 = add_env_binding env3 r t2 env4
			       in Eval steps2 conscase env4 k
			       end
			  else raise FAIL
		       end handle _ => ApplyCont k (mk_term (opid, []) [([],         closure2term (v, e)),
									([],         nilcase),
									([x, xs, r], conscase)],
						    env,
						    num)
		    end
	       else if name = "DoInd"
	       then let val (opid, ([x, rd], downcase), ([], basecase), ([y, ru], upcase), env, k) = dest_DoInd c
		    in let val steps2 = decr_steps num
			   val n = dest_integer v
		       in if n = 0
			  then Eval steps2 basecase env k
			  else if n > 0
			  then let val t2 = mk_term (opid, []) [([],      mk_integer_term (n-1)),
								([x, rd], downcase),
								([],      basecase),
								([y, ru], upcase)]
				   val env1 = add_env_binding env y v e
				   val env2 = add_env_binding env1 ru t2 env
			       in Eval steps2 upcase env2 k
			       end
			  else let val t2 = mk_term (opid, []) [([],      mk_integer_term (n+1)),
								([x, rd], downcase),
								([],      basecase),
								([y, ru], upcase)]
				   val env1 = add_env_binding env x v e
				   val env2 = add_env_binding env1 rd t2 env
			       in Eval steps2 downcase env2 k
			       end
		       end handle _ => ApplyCont k (mk_term (opid, []) [([],      closure2term (v, e)),
									([x, rd], downcase),
									([],      basecase),
									([y, ru], upcase)],
						    env,
						    num)
		    end
	       else if name = "DoLTC"
	       then let val (opid, env, k) = dest_DoLTC c
		    in let val steps2 = decr_steps num
			   val (t2, env2) = limited_type_case2 term v e env
		       in Eval steps2 t2 env2 k
		       end handle _ => ApplyCont k (limited_type_case_no_redex2(**) term v e, env, num)
		    end
	       else if name = "CAN2"
	       then let val (opid, t2, env, k) = dest_CAN2 c
		    in Eval num t2 e (mk_simple_term_op "DoCan" opid [v, e, env, k])
		    end
	       else if name = "DoCan"
	       then let val (opid, v1, e1, env, k) = dest_DoCan c
		    in ApplyCont k (mk_simple_term opid [v1, v], e, num)
		    end
	       else if name = "DoIn"
	       then let val (opid, env, k) = dest_DoIn c
		    in ApplyCont k (mk_simple_term opid [v], e, num)
		    end
	       else raise FAIL
	    end

	and Eval steps term env k =
	    let val op_term = opid_of_term term
	    in if member op_term ["add", "subtract", "multiply", "divide", "remainder"]
	       then let val (opid, [t1, t2]) = dest_simple_term term
		    in Eval steps t1 env (mk_simple_term_op "E2" opid [t2, env, k])
		    end
	       else if op_term = "minus"
	       then let val (opid, [t1]) = dest_simple_term term
		    in Eval steps t1 env (mk_simple_term_op "DoMinus" opid [env, k])
		    end
	       else if member op_term ["less", "int_eq"]
	       then let val (opid, [t1, t2, t3, t4]) = dest_simple_term term
		    in Eval steps t1 env (mk_simple_term_op "C2" opid [t2, t3, t4, env, k])
		    end
	       else if op_term = "atom_eq"
	       then let val (operator, [([], t1),
					([], t2),
					([], t3),
					([], t4)]) = dest_term term
		    in Eval steps t1 env (mk_simple_term_op "AE2" (opid_of_operator operator) [t2, term, t3, t4, env, k])
		    end
	       else if member op_term ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
	       then let val (opid, [t1, t2, t3]) = dest_simple_term term
		    in Eval steps t1 env (mk_simple_term_op "DoTest" opid [t2, t3, env, k])
		    end
	       else if op_term = "spread"
	       then let val (operator, [([], pair), ([x, y], B)]) = dest_term term
		    in Eval steps pair env (mk_term_op "DoSpread" (opid_of_operator operator) [([x, y], B), ([], env), ([], k)])
		    end
	       else if op_term = "decide"
	       then let val (operator, [([], d), ([x], A), ([y], B)]) = dest_term term
		    in Eval steps d env (mk_term_op "DoDecide" (opid_of_operator operator) [([x], A), ([y], B), ([], env), ([], k)])
		    end
	       else if op_term = "apply"
	       then if steps < 0 andalso is_term "ycomb" (subtermn 1 term)
		    (* NOTE: Why do that only in negative-steps mode? *)
		    then let val steps1 = decr_steps steps
			     val f = subtermn 2 term
			 in Eval steps1 (mk_apply_term f term) env k
			 end handle _ => ApplyCont k (term, env(**), steps)
		    else let val (operator, [([], function), ([], arg)]) = dest_term term
			 in Eval steps function env (mk_term_op "DoApply" (opid_of_operator operator) [([], arg), ([], env), ([], k)])
			 end
	       else if op_term = "callbyvalue"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
		    in Eval steps arg env (mk_term_op "DoCallbyvalue" (opid_of_operator operator) [([x], B), ([], env), ([], k)])
		    end
	       else if op_term = "callbyvalueall"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
		    in Evaluator4' steps true arg env (mk_term_op "DoCallbyvalueall" (opid_of_operator operator) [([x], B), ([], env), ([], k)])
		    end
	       else if op_term = "list_ind"
	       then let val (operator, [([], L),
					([], nilcase),
					([x, xs, r], conscase)]) = dest_term term
		    in Eval steps L env (mk_term_op "DoListInd" (opid_of_operator operator) [([], nilcase), ([x, xs, r], conscase), ([], env), ([], k)])
		    end
	       else if op_term = "ind"
	       then let val (operator, [([], i),
					([x, rd], downcase),
					([], basecase),
					([y, ru], upcase)]) = dest_term term
		    in Eval steps i env (mk_term_op "DoInd" (opid_of_operator operator) [([x, rd], downcase), ([], basecase), ([y, ru], upcase), ([], env), ([], k)])
		    end
	       else if op_term = "rec_ind"
	       then let val (operator, [([], arg), ([f, x], B)]) = dest_term term
			val function = mk_lambda_term x (mk_term operator [([], mk_variable_term x), ([f, x], B)])
		    in let val steps1 = decr_steps steps
			   val env1 = add_env_binding env x arg env(**)
			   val env2 = add_env_binding env1 f function env(**)
		       in Eval steps1 B env2 k
		       end handle _ => ApplyCont k (term, env, steps)
		    end
	       else if op_term = "limited_type_case"
	       then Eval steps (subtermn 1 term) env (mk_term_op "DoLTC" op_term [([], env), ([], k)])
	       else if op_term = "TERMOF"
	       then let val steps1 = decr_steps steps
		    in Eval steps1 (apply_conv (TagC (mk_tag_term 1)) term) env k
		    end handle _ => ApplyCont k (term, env(**), steps)
	       else if false (*is_abstraction_term term*)
	       then let val steps1 = decr_steps steps
		    in Eval steps1 (unfold_ab term) env k
		    end handle _ => ApplyCont k (term, env(**), steps)
	       else if op_term = "variable"
	       then let val (a, e) = lookup_binding env (dest_variable term)(**)
		    in Eval steps a e k
		    end handle _ => ApplyCont k (term, env(**), steps)
	       else if member op_term ["pair", "cons"]
	       then if cbv_all
		    then let val (opid, [t1, t2]) = dest_simple_term term
			 in Eval steps t1  env (mk_simple_term_op "CAN2" opid [t2, env, k])
			 end
		    else ApplyCont k (term, env, steps)
	       else if member op_term ["inl", "inr"]
	       then if cbv_all
		    then let val (opid, [t1]) = dest_simple_term term
			 in Eval steps t1 env (mk_simple_term_op "DoIn" opid [env, k])
			 end
		    else ApplyCont k (term, env, steps)
	       else ApplyCont k (term, env, steps)
	    (* NOTE: there is another ite concerning the bound variables of term. *)
	    end
    in Eval steps term env k
    end

fun Evaluator4 steps term =
    Evaluator4' steps false term empty_env halt_term

end
